%%    Eotvos Quantum Transport Utilities - FermiDirac
%    Copyright (C) 2016 Peter Rakyta, Ph.D.
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see http://www.gnu.org/licenses/.
%
%
%> @addtogroup basic Basic Functionalities
%> @{
%> @file FermiDirac.m
%> @brief A class describing the Fermi Dirac distribution of fermionic particles.
%> @} 
%> @brief A class describing the Fermi Dirac distribution of fermionic particles.
%%
classdef FermiDirac < handle
    
    properties (Access = protected)
        %> Boltzmann constant in eV/K
        k_B
        %> The temperature in Kelvin
        T        
        %> 1/(k_B*T)
        beta
        %> treshold temperature (makes difference between T=0 and T>0)
        T_treshold = 1e-10;
        %> The Chemical potential in the same unit as other energy scales in the Hamiltonians.
        mu
    end
    
    
methods (Access=public)
        
%% FermiDirac
%> @brief A constructor of the class
%> @param varargin Cell array of optional parameters (https://www.mathworks.com/help/matlab/ref/varargin.html), for details see #InputParsing.
%> @return Returns with an instance of the class.
    function obj = FermiDirac( varargin )        

        obj.k_B = 8.6173324e-5; % Boltzmann constant in eV/K
        obj.beta = [];

        if strcmpi(class(obj), 'FermiDirac') 
            obj.InputParsing( varargin{:});
        end
        
        
    end
    
    
%% Fermi
%> @brief A function of the Fermi-Dirac statistics
%> @param E The energy value in eV (scalar or an array of energy values). 
%> @return Returns with the occupation numbers (numbers).
    function ret = Fermi(obj, E)
        
        E = reshape(E, 1, numel(E));

		ret = zeros( length(obj.beta), length(E) );
        indexes = logical( abs(obj.T) <= obj.T_treshold);
        
        ret( indexes, E-obj.mu<0 ) = 1;
        ret( indexes ,E-obj.mu==0 ) = 0.5;
        ret( indexes, E-obj.mu>0)  = 0;
        
        if ~isempty(obj.beta(~indexes))
            ret( ~indexes, :) = 1./(1+exp(obj.beta(~indexes)*(E-obj.mu)));
        end
    end  


%% setTemperature
%> @brief Sets the temperature for the calculations. 
%> @param T The temperature in Kelvin (scalar or an array) 
	function setTemperature(obj, T)
        obj.T = reshape( T, numel(T), 1);
		obj.beta = 1./(obj.k_B*T); 
	end
    
    
    
end % public methods
    
    
methods (Access=protected)
       
%% InputParsing
%> @brief Parses the optional parameters for the class constructor.
%> @param varargin Cell array of optional parameters (https://www.mathworks.com/help/matlab/ref/varargin.html):
%> @param 'T' The absolute temperature in Kelvins.
%> @param 'mu' The Chemical potential in the same unit as other energy scales in the Hamiltonians.
    function InputParsing( obj, varargin )
        
        
        p = inputParser;
        p.addParameter('T', 0);
        p.addParameter('mu', 0, @isscalar);
        
        p.parse(varargin{:});
       
        T               = p.Results.T;       
        obj.mu          = p.Results.mu;   
        
        obj.setTemperature(T); 
               

    end        
        
        
end % protected methods

    
end
