%%    Eotvos Quantum Transport Utilities - LoadFromFile
%    Copyright (C) 2009-2016 Peter Rakyta, Ph.D.
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see http://www.gnu.org/licenses/.
%
%> @addtogroup basic Basic Functionalities
%> @{
%> @file LoadFromFile.m
%> @brief A class providing interface to load variables from a file.
%> @} 
%> @brief A class providing interface to load variables from a file.
%%
classdef LoadFromFile < handle

    properties ( Access = protected )
        %> structure containing the loaded data
        LoadedVariable
        %> the filename
        filename
    end
    
methods 
        
%% constructorof the class
%> @brief Constructor of the class.
%> @param filenev The string of the absolute path to the file to open. (Matlab can operate with relative paths as well)
%> @return An instance of the class
function obj = LoadFromFile( filenev )
    
    if exist(filenev, 'file')
        obj.LoadedVariable = load(filenev);
    else
        obj.LoadedVariable = [];
    end
    
    obj.filename = filenev;
            
end

%% LoadVariable
%> @brief Loads a variable from the opened file.
%> @param varname The string containing the name of the variable.
%> @param varargin Cell array of optional parameters (https://www.mathworks.com/help/matlab/ref/varargin.html):
%> @param 'NoEmpty' Set 'On' for throw an error when the loaded variable is empty, or 'Off' (default) otherwise.
%> @param 'StructName' String containing the name of a structure, when the wanted variable is a field of a structure.
function ret = LoadVariable(obj, varname, varargin)
    
    p = inputParser;
    p.addParameter('NoEmpty', 'Off');
    p.addParameter('StructName', []);
    p.parse(varargin{:});      
    
    NoEmpty    = p.Results.NoEmpty;
    StructName = p.Results.StructName;
    
    if ~isempty(StructName)
        dfieldnames = fieldnames(obj.LoadedVariable.(StructName));
        if ismember(varname, dfieldnames)
            ret = obj.LoadedVariable.(StructName).(varname);
        else
            ret = [];
        end
        
        if isempty(ret) && strcmp(NoEmpty, 'On')
           error(['"', obj.filename, '" has no element called "',StructName,'.',varname,'", or it is empty.']);
        end
        
        return;
    end
    
    
    
    if isfield(obj.LoadedVariable, varname)
        ret  = obj.LoadedVariable.(varname);
    else
        ret = [];
    end
    
    if isempty(ret) && strcmp(NoEmpty, 'On')
        error(['"', obj.filename,'" has no element called "',varname,'", or it is empty.']);
    end        
end



%% setLoadedVariables
%> @brief Sets a loaded variable.
%> @param varname The string containing the name of the variable.
%> @param value The value to be set.
%> @param varargin Cell array of optional parameters (https://www.mathworks.com/help/matlab/ref/varargin.html):
%> @param 'NoEmpty' Set 'On' for throw an error when the loaded variable is empty, or 'Off' (default) otherwise.
%> @param 'StructName' String containing the name of a structure, when the wanted variable is a field of a structure.
function ret = setLoadedVariable( obj, varname, value, varargin )
    p = inputParser;
    p.addParameter('NoEmpty', 'Off');
    p.addParameter('StructName', []);
    p.parse(varargin{:});      
    
    NoEmpty    = p.Results.NoEmpty;
    StructName = p.Results.StructName;
    
    if ~isempty(StructName)
        dfieldnames = fieldnames(obj.LoadedVariable.(StructName));
        if ismember(varname, dfieldnames)
            obj.LoadedVariable.(StructName).(varname) = value;
            ret = true;
        else
            ret = false;
        end
        
        if ~ret && strcmp(NoEmpty, 'On')
           error(['"',filenev,'" has no element called "',StructName,'.',varname,'", or it cannot be set.']);
        end
        
        return;
    end
    
    
    
    if isfield(obj.LoadedVariable, varname)
        obj.LoadedVariable.(varname) = value;
        ret = true;
    else
        ret = false;
    end
    
    if ~ret && strcmp(NoEmpty, 'On')
        error(['"',filenev,'" has no element called "',varname,'", or it cannot be set.']);
    end        
end

%% SaveLoadedVariables
%> @brief Save the loaded variables.
%> @param varargin Cell array of optional parameters (https://www.mathworks.com/help/matlab/ref/varargin.html):
%> @param 'filename' A string containing the path to the new file to be created instead of the original one.
function SaveLoadedVariables( obj, varargin )
    p = inputParser;
    p.addParameter('filename', obj.filename);
    p.parse(varargin{:});  
    
    filename = p.Results.filename;
    LoadedVariable = obj.LoadedVariable;
    save( filename, '-struct', 'LoadedVariable' );
end


%% ClearLoadedVariables
%> @brief Disposes the loaded file. 
function ClearLoadedVariables( obj )
    obj.LoadedVariable = [];
end


end %methods


end % Global end
