#include "fintrf.h"
C======================================================================
C     Gateway routine to call the zggev.f function from LAPACK package.
C     (See: http://www.netlib.org/lapack/explore-html/d9/d52/dggev_8f.html for details)
C     Copyright (C) 2009-2015 Peter Rakyta, Ph.D.
C
C     This program is free software: you can redistribute it and/or modify
C     it under the terms of the GNU General Public License as published by
C     the Free Software Foundation, either version 3 of the License, or
C     (at your option) any later version.
C 
C     This program is distributed in the hope that it will be useful,
C     but WITHOUT ANY WARRANTY; without even the implied warranty of
C     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C     GNU General Public License for more details.
C
C     You should have received a copy of the GNU General Public License
C     along with this program.  If not, see http://www.gnu.org/licenses/.
C     
C======================================================================
C     Gateway routine
      subroutine mexFunction(nlhs, plhs, nrhs, prhs)

C     Declarations
      implicit none

      interface
        subroutine getSchur 
     1          (asize, a, ia, ja, ssize, S, iS, jS, mtype)
          INTEGER*4 asize !matrix A is of size asize x asize
          INTEGER*4, ALLOCATABLE :: ia(:), ja(:) 
          REAL*8, ALLOCATABLE ::  a(:)           
          INTEGER*4 ssize !size of the Schur complement to be calculated
          INTEGER, ALLOCATABLE :: iS(:), jS(:) ! row and culomn indices of the Schur complement
          REAL*8, ALLOCATABLE :: S(:) !nonzeros elements in the Schur complement
          INTEGER mtype
        end subroutine
      end interface

C     mexFunction arguments:
      mwPointer plhs(*), prhs(*)
      integer nlhs, nrhs


C     Function declarations:
      mwSize mxGetNzmax
      mwPointer mxGetJc
      mwPointer mxGetIr

C     Arguments for mxCreateNumericArray
      integer*4 complexflag
      mwSize ndim

      mwPointer mxGetPr
      mwPointer mxCreateDoubleMatrix
      mwPointer mxGetM, mxGetN
      integer*4 mxIsDouble
      integer*4 mxIsComplex
      integer*4 mexPrintf

  

C     Array information:
      mwSize nonzerosA !number of nonzeros elements in A
      mwSize sizeAm, sizeAn !size of the matrix A (rows, and cols)  
      integer*4 nonzerosS !number of nonzeros elements in S
      integer*4 sizeS !Size of the Schur complement to be calculated

C     Arguments for computational routine:  
C     fortran representation of the input matrix (CSR format)
      integer*8, allocatable :: ia_int8(:), ja_int8(:) 
      integer*4, allocatable :: ia(:), ja(:)
      real*8, allocatable ::  a(:) 
C     fortran representation of the Schur complement matrix (CSR format)
      integer, allocatable :: iS(:), jS(:) 
      real*8, allocatable ::  S(:) 
      real*8, allocatable ::  S_full(:,:) 
      mwPointer mxS_full
        
C     Parameters for pardiso
      INTEGER mtype
      integer i,j

#ifdef DEBUG
C     Debug parameters
      integer*4 tmp
      character*800 line
#endif


C-----------------------------------------------------------------------
C     Check for proper number of arguments. 
      if(nrhs .ne. 2) then
         call mexErrMsgIdAndTxt ('MATLAB:timestwo:nInput',
     +                           'Two inputs required.')
      elseif(nlhs .gt. 1) then
         call mexErrMsgIdAndTxt ('MATLAB:timestwo:nOutput',
     +                           'Too many output arguments.')
      endif

C     Validate inputs
      
C     Check that the input is a double.
      if ((mxIsDouble(prhs(1)) .eq. 0)) then
         call mexErrMsgIdAndTxt ('MATLAB:timestwo:NonDouble',
     +                           'Input must be a double.')
      endif
      
C     Check if inputs are complex.
      if ((mxIsComplex(prhs(1)) .ne. 0) ) then
         call mexErrMsgIdAndTxt ('MATLAB:convec:NonComplex',
     +                           'Inputs must be real.')
      endif    

C     Get the size of the input array. (MATLAB)
      nonzerosA = mxGetNzmax(prhs(1))
      sizeAm = mxGetM(prhs(1))
      sizeAn = mxGetN(prhs(1))


C     Allocating fortran arrays
      allocate(ia_int8(sizeAn+1))
      allocate(ja_int8(nonzerosA))
      allocate(ia(sizeAn+1))
      allocate(ja(nonzerosA))
      allocate(a(nonzerosA))
C
C     Check the allocated arrays
      if (.not. ( allocated(ia_int8) .and.
     + allocated(ja_int8) .and.
     + allocated(ia) .and.
     + allocated(ja) .and.
     + allocated(a) )) then
         call mexErrMsgIdAndTxt ('MATLAB:dgetSchur:BadAllocation',
     +       'The allocation of variables was unsuccesfull.')
      endif
C     Copy data into the fortran arrays
C     MATLAB does CSC format and PARDISO CSR format for the sparse matrices! Need to transpose.   
      call mxCopyPtrToInteger8( 
     +                   mxGetJc(prhs(1)), ia_int8, sizeAn+1 )
      call mxCopyPtrToInteger8( 
     +                   mxGetIr(prhs(1)), ja_int8, nonzerosA )
      call mxCopyPtrToReal8( mxGetPr(prhs(1)), a, nonzerosA )
C     Fortran starts the indexes from 1
      ia = int(ia_int8, 4) + 1
      ja = int(ja_int8, 4) + 1
      deallocate(ia_int8)
      deallocate(ja_int8)

C     choose the matrix type for pardiso factorization
      mtype     = 1  ! structurally symmetric real matrix
C     getting the size of the Schur complement
      ndim = 1
      call mxCopyPtrToInteger4( mxGetPr(prhs(2)), sizeS, ndim )


C     Do the pardiso calulations to get the Schur complement
C     Passing deallocated arrays might cause troubles
      ALLOCATE(iS(0))
      ALLOCATE(jS(0))
      ALLOCATE(S(0))
      CALL getSchur(int(sizeAm, 4), a, ia, ja, 
     +   sizeS, S, iS, jS, mtype)

C     deallocate the fortran matrix A
      deallocate(ia)
      deallocate(ja)
      deallocate(a)


#ifdef DEBUG
      write(line,*) NEW_LINE('A')
      tmp = mexPrintf(line)
      write(line,*) 'The Calculated Schur complement:', NEW_LINE('A')
      tmp = mexPrintf(line)
      i = 0
      do j = 1, size(S)
        if (j >= iS(i+1)) then
          i=i+1
        end if

        if (j < iS(i+1)) then  
          write(line,*) 'row', i, 'col', jS(j),
     + 'element of S = ', S(j), NEW_LINE('A')
          tmp = mexPrintf(line)
        end if
      end do

#endif

C     creating full S matrix
      ALLOCATE(S_full(sizeS, sizeS))
      S_full = 0.d0
      i = 0
      do j = 1, size(S)
        if (j >= iS(i+1)) then
          i=i+1
        end if

        if (j < iS(i+1)) then 
C     MATLAB does CSC format and PARDISO CSR format for the sparse matrices! Need to transpose back.          
          S_full(jS(j), i) = S(j)
        end if
      end do
      ndim = int(sizeS, 8)
      complexflag = 0;
      plhs(1) = mxCreateDoubleMatrix(ndim, ndim, 
     + complexflag) ! for S_full
      mxS_full = mxGetPr(plhs(1))
      call mxCopyReal8ToPtr(S_full, mxS_full,ndim*ndim) 




      
C.. Deallocating memory for the Schur complement
      if ( allocated(iS) ) then
        DEALLOCATE(iS)
      endif

      if ( allocated(jS) ) then
        DEALLOCATE(jS)
      endif

      if ( allocated(S) ) then
        DEALLOCATE(S)
      endif     

      if ( allocated(S_full) ) then
        DEALLOCATE(S_full)
      endif      


      return
      end


