#include "fintrf.h"
C======================================================================
C     Gateway routine to call the zggev.f function from LAPACK package.
C     (See: http://www.netlib.org/lapack/explore-html/d9/d52/dggev_8f.html for details)
C     Copyright (C) 2009-2015 Peter Rakyta, Ph.D.
C
C     This program is free software: you can redistribute it and/or modify
C     it under the terms of the GNU General Public License as published by
C     the Free Software Foundation, either version 3 of the License, or
C     (at your option) any later version.
C 
C     This program is distributed in the hope that it will be useful,
C     but WITHOUT ANY WARRANTY; without even the implied warranty of
C     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C     GNU General Public License for more details.
C
C     You should have received a copy of the GNU General Public License
C     along with this program.  If not, see http://www.gnu.org/licenses/.
C     
C======================================================================
C     Gateway routine
      subroutine mexFunction(nlhs, plhs, nrhs, prhs)

C     Declarations
      implicit none

C     mexFunction arguments:
      mwPointer plhs(*), prhs(*)
      integer nlhs, nrhs


C     Function declarations:
      mwSize mxGetNzmax
      mwPointer mxGetJc
      mwPointer mxGetIr
      mwPointer mxCreateString
      mwPointer mxGetString, getstring_stat 
      integer mxIsChar

C     Arguments for copying data between fortran and MATLAB
      integer*4 complexflag

      mwPointer mxGetPr, mxGetPi
      mwPointer mxCreateDoubleMatrix
      mwPointer mxGetM, mxGetN
      integer*4 mxIsDouble
      integer*4 mxIsComplex
      integer*4 mexPrintf

      logical fileExist
      character(len=100) :: filename 
      character(len=7) :: file_status

C     Array information:
      mwSize nonzerosA !number of nonzeros elements in A
      mwSize sizeAm, sizeAn !size of the matrix A (rows, and cols) 
      mwPointer mrows, ncols !size of the input string containing the filename
      integer*4 strlen ! the length of the input string
      integer*4 sizeS !Size of the Schur complement to be calculated 

C     Arguments for computational routine:  
C     fortran representation of the input matrix (CSR format)
      integer*8, allocatable :: ia_int8(:), ja_int8(:) 
      integer*4, allocatable :: ia(:), ja(:)
      complex*16, allocatable ::  a(:) 
        
C     Parameters for pardiso
      integer i,j

#ifdef DEBUG
C     Debug parameters
      integer*4 tmp
      character*800 line
#endif


C-----------------------------------------------------------------------
C     Check for proper number of arguments. 
      if(nrhs .ne. 2) then
         call mexErrMsgIdAndTxt ('MATLAB:export_real_sparse:nInput',
     +                           'Two inputs required.')
      elseif(nlhs .gt. 0) then
         call mexErrMsgIdAndTxt ('MATLAB:export_real_sparse:nOutput',
     +                           'Too many output arguments.')
      endif

C     Validate inputs
      
C     Check that the input is a double.
      if ((mxIsDouble(prhs(1)) .eq. 0)) then
         call mexErrMsgIdAndTxt ('MATLAB:export_real_sparse:NonDouble',
     +                           'Input must be a double.')
      endif
      
C     Check if input matrix must be complex
      if ((mxIsComplex(prhs(1)) .ne. 1) ) then
         call mexErrMsgIdAndTxt ('MATLAB:convec:NonComplex',
     +                           'Inputs must be complex.')
      endif    

C     The input filename must be a string.
       if(mxIsChar(prhs(2)) .ne. 1) then
         call mexErrMsgIdAndTxt ('MATLAB:export_real_sparse:NonString',
     +                 'Input filename must be a string.')
       endif

C     Get the size of the input string.
       mrows = mxGetM(prhs(2))
       ncols = mxGetN(prhs(2))

C     Get the length of the input string and validate.
       strlen = int(mrows*ncols, 4)
       if (strlen .gt. len(filename)) then
         call mexErrMsgIdAndTxt ('MATLAB:export_real_sparse:NonDouble',
     +             'input filename is greater than max str size.')   
       endif

C     Get the input filename.
      getstring_stat = 
     + mxGetString(prhs(2), filename, int(len(filename), 8))

#ifdef DEBUG
      write(line,*) 'Input file name: ',filename, NEW_LINE('A')
      tmp = mexPrintf(line)
#endif


C     Get the size of the input array. (MATLAB)
      nonzerosA = mxGetNzmax(prhs(1))
      sizeAm = mxGetM(prhs(1))
      sizeAn = mxGetN(prhs(1))


C     Allocating fortran arrays
      allocate(ia_int8(sizeAn+1))
      allocate(ja_int8(nonzerosA))
      allocate(ia(sizeAn+1))
      allocate(ja(nonzerosA))
      allocate(a(nonzerosA))
C
C     Check the allocated arrays
      if (.not. ( allocated(ia_int8) .and.
     + allocated(ja_int8) .and.
     + allocated(ia) .and.
     + allocated(ja) .and.
     + allocated(a) )) then
         call mexErrMsgIdAndTxt ('MATLAB:dgetSchur:BadAllocation',
     +       'The allocation of variables was unsuccesfull.')
      endif
C     Copy data into the fortran arrays
C     MATLAB does CSC format and PARDISO CSR format for the sparse matrices! Need to transpose.   
      call mxCopyPtrToInteger8( 
     +                   mxGetJc(prhs(1)), ia_int8, sizeAn+1 )
      call mxCopyPtrToInteger8( 
     +                   mxGetIr(prhs(1)), ja_int8, nonzerosA )
      call mxCopyPtrToComplex16( mxGetPr(prhs(1)), mxGetPi(prhs(1)), 
     +  a, nonzerosA )
C     Fortran starts the indexes from 1
      ia = int(ia_int8, 4) + 1
      ja = int(ja_int8, 4) + 1
      deallocate(ia_int8)
      deallocate(ja_int8)



C     Opening the file to export data 
      inquire(file=trim(filename),exist=fileExist)!check if it exist
      if(.not.fileExist)then
        file_status = 'NEW'
      else
        file_status = 'REPLACE'
      endif   
      OPEN(UNIT=6, FILE=trim(filename),STATUS=file_status
     + ,FORM='UNFORMATTED')

      WRITE(6) int(sizeAn, 4)
      WRITE(6) int(nonzerosA, 4)
      WRITE(6) ia(:)
      WRITE(6) ja(:)
      WRITE(6) a(:)


C     Closing the file
      CLOSE(UNIT=6)

C     deallocate the fortran matrix A
      deallocate(ia)
      deallocate(ja)
      deallocate(a)

      return
      

      end


