!======================================================================
!     Wrapper module to read in the CMLComp XML file generated by the SIESTA package.
!     This file is part of the EQuUs package. (for details see: http://eqt.elte.hu/equus/home)
!     Copyright (C) 2016 Peter Rakyta, Ph.D.
!
!     This program is free software: you can redistribute it and/or modify
!     it under the terms of the GNU General Public License as published by
!     the Free Software Foundation, either version 3 of the License, or
!     (at your option) any later version.
! 
!     This program is distributed in the hope that it will be useful,
!     but WITHOUT ANY WARRANTY; without even the implied warranty of
!     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!     GNU General Public License for more details.
!
!     You should have received a copy of the GNU General Public License
!     along with this program.  If not, see http://www.gnu.org/licenses/.
!     
!======================================================================
#ifndef MAX_PROP_LENGTH
#define MAX_PROP_LENGTH 100
#endif



MODULE Fox_sax_EQuUs

  use EQuUs_C
  use FoX_sax
  use EQuUs_XML

! Derived type holding the XML parser 
  type(xml_t), private :: xp
! The number of the retrived nodes
  integer :: numTargetNodes
! The array of the retrived XML target nodes
  TYPE(XMLnode), pointer :: TargetNodes(:)
! The representation of the current XML target node in focus
  TYPE(XMLnode), pointer :: currentTargetNode
! The representation of the XML node in focus
  TYPE(XMLnode), pointer :: currentNode
! Logical value. True if a desired XML node is in focus
  logical, private :: targetNodeActive = .false.
! The level of the current node in the XML tree
  integer, private :: subnodelevel = 0
! The local names of the properties to be retrived from the XML
  character(len=MAX_PROP_LENGTH, kind=C_CHAR), pointer :: property_names(:)
 


!****************************************************************      
  CONTAINS
!****************************************************************




!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file hsx_EQuUs.f90
!> @brief Converts an array of XML subnodes into an EQstruct structure prototypes.
!! 
!> @param[out] subnodes An EQstruct structure prototype.
!> @param[in] XMLsubnodes The XML nodes to be converted into EQstruct structure prototypes.
  subroutine convertsubnodes2eq_struct( subnodes, XMLsubnodes )

    TYPE(EQstruct), pointer :: subnodes
    TYPE(XMLnode), pointer :: XMLsubnodes(:)

    TYPE(EQstruct), pointer :: eq_struct => null()
    TYPE(EQelement), pointer :: eq_elements(:)
    integer nodenum, idx

#ifdef DEBUG
    print *, "Converting XML subnodes to eq_structs"
#endif

    nodenum = SIZE( XMLsubnodes, 1)


    allocate( eq_elements(nodenum) )
    do idx = 1, nodenum
      call convert2eq_struct( eq_struct, XMLsubnodes(idx) )        
#ifdef DEBUG
      print *, "localname of the current subnode: ", trim(XMLsubnodes(idx)%localname), " numfields: ", eq_struct%numfields
#endif

      call assocElement( eq_elements(idx), "subnode", eq_struct )

    end do

    allocate(subnodes)
    subnodes%fields = C_LOC(eq_elements(1))
    subnodes%numfields = nodenum

#ifdef DEBUG
    print *, "Converting XML subnodes to eq_structs ------- finished"
#endif


  end subroutine convertsubnodes2eq_struct



!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file hsx_EQuUs.f90
!> @brief Converts an XML node into an EQstruct structure prototype.
!! 
!> @param[out] eq_struct The EQstruct structure prototype.
!> @param[in] node The XML node to be converted into EQstruct structure prototype.
  subroutine convert2eq_struct( eq_struct, node )

    TYPE(EQstruct), pointer :: eq_struct
    TYPE(XMLnode) :: node
    TYPE(EQelement), pointer :: eq_elements(:)
    TYPE(EQstruct), pointer :: subnodes => null()
    character(len=:), pointer :: value
    integer elementnum, idx

#ifdef DEBUG
    print *, "Converting XML node to eq_struct"
#endif   

    ! converting subnodes/value
    if ( hasSubnodes(node) ) then
        call convertsubnodes2eq_struct( subnodes, node%subnodes )
    end if


    ! first eq_element stores the localname
    elementnum = 1

    if (ASSOCIATED(node%attrs)) then
      elementnum = elementnum + SIZE( node%attrs, 1 )
    end if

    if ((ASSOCIATED(node%subnodes)).or.(ASSOCIATED(node%value))) then
      elementnum = elementnum + 1
    end if


    if (elementnum.eq.0 ) then
      return
    end if

    allocate( eq_elements(elementnum) )
     

    ! converting the localname
    allocate( character(len=LEN(node%localname)) :: value )
    value = node%localname
    call assocElement( eq_elements(1), "localname", value )
    deallocate(value)


    ! converting atrributes
    if (ASSOCIATED(node%attrs)) then
      do idx = 1, SIZE( node%attrs, 1 )
        allocate( character(len=LEN(node%attrs(idx)%value)) :: value )
        value = node%attrs(idx)%value
        call assocElement( eq_elements(idx+1), node%attrs(idx)%name, value )
        deallocate(value)
      end do
    end if


    ! converting subnodes/value
    if ( hasSubnodes(node) ) then
        call assocElement( eq_elements(elementnum), "subnodes", subnodes )

    else if (ASSOCIATED(node%value)) then
        allocate( character(len=LEN(node%value)) :: value )
        value = node%value
        call assocElement( eq_elements(elementnum), "value", value )
        deallocate(value)
    end if

#ifdef DEBUG
    print *, "allocate ", trim(node%localname)
#endif
    allocate( eq_struct )
    eq_struct%fields = C_LOC(eq_elements(1))
    eq_struct%numfields = elementnum

#ifdef DEBUG
    print *, "Converting XML node to eq_struct ------ finished"
#endif

  end subroutine convert2eq_struct






!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file hsx_EQuUs.f90
!> @brief Gets the retrived XML node converted into EQstruct structure prototype.
!! 
!> @param[out] eq_struct The EQstruct structure prototype.
!> @param[in] idx The index of the retrived XML node to be converted into EQstruct structure prototype.
  subroutine get_eq_structure( eq_struct, idx )

    TYPE(EQstruct) :: eq_struct
    TYPE(EQstruct), pointer :: peq_struct
    integer idx

    if ( idx.gt.numTargetNodes ) then
      print *, "maximal value for idx is: ", numTargetNodes
      return
    end if

    call convert2eq_struct( peq_struct, TargetNodes(idx) )
    eq_struct = peq_struct

  end subroutine get_eq_structure


!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file hsx_EQuUs.f90
!> @brief read in particular nodes from the CMLComp XML into EQuUs structure prototype
!! 
!> @param[in] fname The name of the file to be loaded
!> @param[in] fname_len The length of the filename
!> @param[out] eq_struct An array of EQuUs structure prototypes corresponding to the desired properties
! 
  subroutine read_xml( fname, fname_len, cproperty_names, numofProperties)


  integer*8 fname_len
  character(len=fname_len), intent(in) :: fname
  logical :: file_exist
  integer numofProperties
  TYPE(C_PTR) :: cproperty_names(numofProperties)
  character(len=MAX_PROP_LENGTH, kind=C_CHAR), pointer :: property_name

  TYPE(EQstruct) :: XML_node_tmp

  integer idx


  


! double check for the existence of the file 
  inquire(file=fname, exist=file_exist)
  if (.not.file_exist) then
    print *, 'File does not exist'
    stop
  endif

! Open the XML file
  call open_xml_file(xp, fname, iostat=iostat)
  if (iostat /= 0) then
    write(*,*) "Cannot open file."
    stop
  endif


  allocate( property_names(numofProperties))
  do idx = 1, numofProperties
    call c_f_pointer( cproperty_names(idx), property_name )
    property_names(idx) = property_name
  end do

#ifdef DEBUG
  print *, "Property names to be retrived:"
  do idx = 1, numofProperties
    print *, trim(property_names(idx))
  end do
#endif

  call parse(xp, startElement_handler=startElement_handler, &
                  endElement_handler = endElement_handler, &
                  error_handler=error_handler, &
                  fatalError_handler = fatalError_handler, &
                  characters_handler = characters_handler)

! Close the XML file
  call close_xml_t(xp)

! deallocating the property names
  deallocate( property_names )

#ifdef DEBUG
  print *, NEW_LINE('A'), NEW_LINE('A'), NEW_LINE('A')
  print *, "The retrived nodes:"
  do idx = 1, SIZE( TargetNodes, 1)
    call printnodeWithSubnodes( TargetNodes(idx) )
    print *, NEW_LINE('A'), NEW_LINE('A')
  end do
#endif



  end subroutine read_xml



!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file Fox_sax_EQuUs.f90
!> @brief Deallocates the loaded target nodes
!! 
  subroutine deallocateTargetNodes()

  integer idx

  do idx = 1, SIZE( TargetNodes, 1)
    call deallocateNode(TargetNodes(idx))
  end do
  deallocate( TargetNodes )

  end subroutine deallocateTargetNodes





!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file Fox_sax_EQuUs.f90
!> @brief Add an XML node to the array of the retrived XML target nodes
!! 
!> @param[in] node A pointer to a retrived XML node.
  subroutine addTargetNode( node )
    TYPE(XMLnode), pointer :: node
    TYPE(XMLnode), pointer :: nodes_tmp(:)
    integer nodenum, idx

    if (.not.ASSOCIATED(TargetNodes)) then
      allocate(TargetNodes(1))
      call copynode(currentTargetNode, TargetNodes(1))
      numTargetNodes = 1
      return
    end if


    nodenum = SIZE( TargetNodes, 1 )
    allocate( nodes_tmp(nodenum) )

    do idx = 1, nodenum
      call copynode(TargetNodes(idx), nodes_tmp(idx) )
    end do

    deallocate( TargetNodes )
    allocate( TargetNodes( nodenum+1 ) )
    do idx = 1, nodenum
      call copynode( nodes_tmp(idx), TargetNodes(idx) )
    end do 
    deallocate( nodes_tmp )

    call copynode( node, TargetNodes(nodenum+1) ) 

    numTargetNodes = numTargetNodes + 1

  end subroutine addTargetNode

!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file Fox_sax_EQuUs.f90
!> @brief Check whether the XML node in focus is about to retrive or not.
!! 
!> @param[in] chunk The local name, or an attribute value of the XML node.
!> @return Returns true if the XML node in focus is about to retrive, or false otherwise.
  logical function checkname( chunk )
    character(len=*), intent(in) :: chunk
    integer idx

    do idx = 1, SIZE( property_names, 1)
      if ( chunk == property_names(idx) ) then
        checkname = .true.
        return
      end if
    end do

    checkname = .false.
    return
    
  end function checkname


!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file Fox_sax_EQuUs.f90
!> @brief error_handler function defined in the FoX_sax package (http://www1.gly.bris.ac.uk/~walker/FoX/DoX/FoX_sax.html)
!! 
!> @param[in] msg The error message.
  subroutine error_handler(msg)
    character(len=*), intent(in) :: msg
    print *,  "Error encountered and caught"
    print *,  msg
  end subroutine error_handler

!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file Fox_sax_EQuUs.f90
!> @brief fatalError_handler function defined in the FoX_sax package (http://www1.gly.bris.ac.uk/~walker/FoX/DoX/FoX_sax.html)
!! 
!> @param[in] msg The error message.
  subroutine fatalError_handler(msg)
    character(len=*), intent(in) :: msg
    print *,  "Fatal error encountered and caught"
    print *,  msg
  end subroutine fatalError_handler


!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file Fox_sax_EQuUs.f90
!> @brief characters_handler function defined in the FoX_sax package (http://www1.gly.bris.ac.uk/~walker/FoX/DoX/FoX_sax.html)
!! 
!> @param[in] chunk The string value of the XML node.
  subroutine characters_handler(chunk)
    character(len=*), intent(in) :: chunk

    character(len=len(chunk)+1), pointer :: value
    

    if (.not.targetNodeActive) then
      return
    endif

! XML node containing subnodes should not contain any value
    if (ASSOCIATED(currentNode%subnodes)) then
      currentNode%value => null()
      return
    end if

    allocate( character(len=len(chunk)+1) :: currentNode%value )
    currentNode%value = trim(chunk)
#ifdef DEBUG
    print *, "assigning value to the currentNode: ", trim(currentNode%value)
#endif   




    
  end subroutine characters_handler

!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file Fox_sax_EQuUs.f90
!> @brief endElement_handler function defined in the FoX_sax package (http://www1.gly.bris.ac.uk/~walker/FoX/DoX/FoX_sax.html)
!! 
!> @param[in] URI 
!> @param[in] localname 
!> @param[in] name  
  subroutine endElement_handler(URI, localname, name)
    character(len=*), intent(in)     :: URI
    character(len=*), intent(in)     :: localname
    character(len=*), intent(in)     :: name


    if ( targetNodeActive ) then
      ! the target node is going out of scope
      if ( subnodelevel.eq.currentTargetNode%subnodelevel ) then
#ifdef DEBUG
print *, "Deactivatig the current target node"
#endif
        targetNodeActive = .false.
        call addTargetNode( currentTargetNode )
        currentTargetNode => null()
        currentNode => null()
      end if

    else
        subnodelevel = subnodelevel -1 
        return
    end if



    if (ASSOCIATED(currentNode) ) then
      if (ASSOCIATED(currentNode%parentnode)) then
        call addsubnode( currentNode%parentnode, currentNode )
        currentNode => currentNode%parentnode
      end if
    end if

    subnodelevel = subnodelevel -1 

  end subroutine endElement_handler


!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file Fox_sax_EQuUs.f90
!> @brief startElement_handler function defined in the FoX_sax package (http://www1.gly.bris.ac.uk/~walker/FoX/DoX/FoX_sax.html)
!! 
!> @param[in] URI The URL related to the XML node
!> @param[in] localname The localname of the XML node
!> @param[in] name The name of the XML node
!> @param[in] attributes Dictionary conatining the attributes of an XML node
  subroutine startElement_handler(URI, localname, name,attributes)
    character(len=*), intent(in)   :: URI  
    character(len=*), intent(in)   :: localname
    character(len=*), intent(in)   :: name 
    type(dictionary_t), intent(in) :: attributes


    type(XMLnode), pointer :: parentnode

    integer :: idx
    character(len=MAX_PROP_LENGTH), pointer :: XML_attr_value(:)

    TYPE( ATTRIB ), pointer :: attrs(:)
    type(XMLnode), pointer :: subnode

! The number of XML attributes in a given node
    integer :: XML_attributes_num

    subnodelevel = subnodelevel + 1

    if ( .not.targetNodeActive ) then
      call createTargetNode( URI, localname, name,attributes )
      currentNode => currentTargetNode
      return
    end if


    ! create subnode in the current target node
#ifdef DEBUG
print *, "allocating subnode: ", localname
#endif

    allocate( subnode )
    XML_attributes_num = getLength(attributes)

    subnode%subnodelevel = subnodelevel
    subnode%localname = trim(localname)
    subnode%subnodes=> null() 
    subnode%value=> null() 
    allocate( subnode%attrs(XML_attributes_num) )

    do idx = 1, getLength(attributes)
      subnode%attrs(idx)%name = trim(getQName(attributes, idx))
      subnode%attrs(idx)%value = trim(getValue(attributes, idx))
    enddo

    
    subnode%parentnode => currentNode
    currentNode => subnode

  end subroutine startElement_handler



!-------------------------------------------------------------------------------------------------
! DESCRIPTION:
!> @file Fox_sax_EQuUs.f90
!> @brief Check whether the given XML node is one of the target nodes and creates a target node.
!! 
!> @param[in] URI The URL related to the XML node
!> @param[in] localname The localname of the XML node
!> @param[in] name The name of the XML node
!> @param[in] attributes Dictionary conatining the attributes of an XML node
  subroutine createTargetNode(URI, localname, name,attributes)
    character(len=*), intent(in)   :: URI  
    character(len=*), intent(in)   :: localname
    character(len=*), intent(in)   :: name 
    type(dictionary_t), intent(in) :: attributes

! The number of XML attributes in a given node
    integer :: XML_attributes_num

    integer :: idx
    character(len=MAX_PROP_LENGTH), pointer :: XML_attr_value(:)

    TYPE( ATTRIB ), pointer :: attrs(:)
     
    ! retriving the attributes of a desired property
    if ( (localname.EQ."property").OR.                                     &
       (localname.EQ."propertyList") ) then
      do idx = 1, getLength(attributes)
        if ( (getQName(attributes, idx)=="dictRef").OR.                    &
           (getQName(attributes, idx)=="title")) then
          if (checkname( getValue(attributes, idx) )) then 
            XML_attributes_num = getLength(attributes)
            targetNodeActive = .true.
          endif
        endif
      enddo

      if ( .not.targetNodeActive ) then
        return
      end if 
#ifdef DEBUG
print *, "Allocating target node"
#endif
      allocate(currentTargetNode)
      currentTargetNode%subnodelevel = subnodelevel
      currentTargetNode%localname = trim(localname)
      currentTargetNode%subnodes=> null()
      currentTargetNode%parentnode=> null()
      currentTargetNode%value=> null()
      allocate( currentTargetNode%attrs(XML_attributes_num) )

      do idx = 1, getLength(attributes)
        currentTargetNode%attrs(idx)%name = trim(getQName(attributes, idx))
        currentTargetNode%attrs(idx)%value = trim(getValue(attributes, idx))
      enddo

      return
    end if


  end subroutine createTargetNode




end module


