/**======================================================================
     Gateway routine to read in the CMLComp XML document generated by the SIESTA package.
     Copyright (C) 2016 Peter Rakyta, Ph.D.

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see http://www.gnu.org/licenses/.
     
======================================================================*/
#include "mex.h"
#include <unistd.h>
#include <string.h>
#include "EQuUs_MATLAB.h"

#ifndef MAX_PROP_LENGTH
#define MAX_PROP_LENGTH 100
#endif

#ifdef __cplusplus
extern "C" {
#endif


#ifdef __INTEL_COMPILER
void fox_sax_equus_mp_read_xml_( char* fnamget_eq_structurese, long *fname_len, char** property_names, int* numodProperties );
void fox_sax_equus_mp_get_eq_structure_( EQstruct* eq_struct , int* idx);
void fox_sax_equus_mp_deallocatetargetnodes_();
int fox_sax_equus_mp_numtargetnodes_;
#else
void __fox_sax_equus_MOD_read_xml( char* fnamget_eq_structurese, long *fname_len, char** property_names, int* numodProperties );
void __fox_sax_equus_MOD_get_eq_structure( EQstruct* eq_struct , int* idx);
void __fox_sax_equus_MOD_deallocatetargetnodes();
int __fox_sax_equus_MOD_numtargetnodes;
#endif  

#ifdef __cplusplus
 }
#endif

/** Routine to turn nested EQstruct structure prototypes into MATLAB structure.
 \param eq_struct A pointer to an EQstruct structure prototype.
 \param Mstruct	A pointer to a MATLAB structure.
*/
void XMLnode2MatlabStruct( EQstruct* eq_struct, mxArray* Mstruct );


/** Gateway routine to read in the hsx structure generated by the SIESTA package (see http://www.mathworks.com/help/matlab/matlab_external/standalone-example.html#zmw57dd0e18558).
 \param nlhs	Number of output (left-side) arguments, or the size of the plhs array.
 \param plhs	Array of output arguments.
 \param nrhs	Number of input (right-side) arguments, or the size of the prhs array.
 \param prhs	Array of input arguments.
*/
void mexFunction(int nlhs, mxArray *plhs[],
                 int nrhs, const mxArray *prhs[])
{
      /* input filename */
      char *filename;
      /* length of the filename */
      long buflen;
      /* field names of the output structure */
      char **fieldnames; 
      /* names of the properties to be retrived from the XML */ 
      char** property_names;
      /* number of the properties to be retrived from the XML */ 
      int numofProperties;
      /* NUmber of obtained target XML nodes */
      int numTargetNodes;

/*-----------------------------------------------------------------------*/
/*     Check for proper number of arguments. */

      if(nrhs != 2) {
         mexErrMsgIdAndTxt ("siesta_EQuUs:read_SiestaXML:nInput", "Two inputs required.");
      }
      else if (nlhs > 1) {
         mexErrMsgIdAndTxt ("siesta_EQuUs:read_SiestaXML:nOutput", "Too many output arguments.");
      }

      /* Validate inputs */
     
      /* Check that the first input is a string.*/
      if ((mxIsChar(prhs[0]) != 1)) {
         mexErrMsgIdAndTxt ("siesta_EQuUs:read_SiestaXML:NonString", "First input must be a string.");
      }


      /* copy the string data from prhs[0] into a C string.    */
      filename = mxArrayToString(prhs[0]);
#ifdef DEBUG
      mexPrintf( "%s\n", filename );
#endif

      if (access(filename, F_OK) == -1) {
        mexErrMsgIdAndTxt ("siesta_EQuUs:read_SiestaXML:FileNotExist", "The given file %s does not exist.", filename);
      }



      /* Check that the second input is a cell of strings.*/
      int idx;
      mxArray* cell;
      if ((mxIsCell(prhs[1]) != 1)) {
        mexErrMsgIdAndTxt ("siesta_EQuUs:read_SiestaXML:NonCell", "Second input must be a cell.");
      }
      if ( (mxGetN(prhs[1]) >1) && (mxGetM(prhs[1]) >1) ) {
        mexErrMsgIdAndTxt ("siesta_EQuUs:read_SiestaXML:NonVector", "Second input must be a cell vector.");
      }
      numofProperties = mxGetN(prhs[1]) * mxGetM(prhs[1]);
      property_names = (char**)malloc( numofProperties*sizeof(char*) );
      for ( idx = 0; idx<numofProperties; idx++ ) {
        cell = mxGetCell( prhs[1], idx);
        if ( mxGetN(cell)*mxGetM(cell)+1 > MAX_PROP_LENGTH ) {
          mexErrMsgIdAndTxt ("siesta_EQuUs:read_SiestaXML:MAX_PROP_LENGTH", "Increase the maximal length of the proeprty names.");
        }
        if (!mxIsChar(cell)) {
          mexErrMsgIdAndTxt ("siesta_EQuUs:read_SiestaXML:NonString", "Second input must contain of strings.");
        }
        property_names[idx] = (char*)malloc( MAX_PROP_LENGTH*sizeof(char) );
        buflen = (mxGetM(cell) * mxGetN(cell)) + 1;
        mxGetString(cell, property_names[idx], MAX_PROP_LENGTH);
        convert2FortranString( property_names[idx], MAX_PROP_LENGTH );
      }



      /* get the length of the input file */
      buflen = (mxGetM(prhs[0]) * mxGetN(prhs[0])) + 1;

      /* Call the fortran routines to parse the XML file */
#ifdef __INTEL_COMPILER
      fox_sax_equus_mp_read_xml_( filename, &buflen, property_names, &numofProperties); 
      numTargetNodes = fox_sax_equus_mp_numtargetnodes_;
#else
       __fox_sax_equus_MOD_read_hsx( filename, &buflen, eq_structs);
      numTargetNodes = __fox_sax_equus_MOD_numtargetnodes;
#endif  


#ifdef DEBUG
      mexPrintf( "Number of obtained target nodes: %d\n", numTargetNodes);
#endif 


      /* creating en empty structure for MATLAB */
      plhs[0] = mxCreateStructMatrix(1, 1, 0, NULL);


      /* retriving the obtained XML nodes */
      EQstruct eq_struct;
      for( idx=1; idx<=numTargetNodes; idx++ ) {

#ifdef __INTEL_COMPILER
      fox_sax_equus_mp_get_eq_structure_( &eq_struct, &idx );
#else
      __fox_sax_equus_MOD_get_eq_structure( eq_struct, &idx );
#endif  

      /* regularize the obtained structure */
      regularizeEQstruct( &eq_struct );
//continue;
      /* adding XML node to the output structure */
      XMLnode2MatlabStruct( &eq_struct, plhs[0] );


      }

#ifdef DEBUG
      mexPrintf( "\nDeallocating Target nodes\n*********************************\n" );
#endif

#ifdef __INTEL_COMPILER
      fox_sax_equus_mp_deallocatetargetnodes_();
#else
      __fox_sax_equus_MOD_deallocatetargetnodes();
#endif  
    
}


/** Routine to turn nested EQstruct structure prototypes into MATLAB structure.
 \param eq_struct A pointer to an EQstruct structure prototype.
 \param Mstruct	A pointer to a MATLAB structure.
*/
void XMLnode2MatlabStruct( EQstruct* eq_struct, mxArray* Mstruct ) {


  int idx;
  EQelement* field;
  mxArray* struct_tmp;
  char** localname = NULL;

  if (mxIsStruct( Mstruct ) == 0) {
        mexErrMsgIdAndTxt ("read_SiestaXML:XMLnode2MatlabStruct:NotStructure", "The given mxArray is not a structure.");
  }

  struct_tmp = mxCreateStructMatrix(1, 1, 0, NULL);

  for( idx=0; idx<=eq_struct->numfields-1; idx++ ) {
    field = &eq_struct->fields[idx];
    if (( strcmp(field->elementname, "localname") == 0) && localname==NULL) {
      localname = (char**)field->value;
    }
    else if ( strcmp(field->elementname, "subnodes") == 0) {
      EQstruct* subnodes = (EQstruct*)field->value;
      EQstruct* subnode;
      mxArray* Msubnode = mxCreateStructMatrix(1, 1, 0, NULL);
#ifdef DEBUG
      mexPrintf("number of subnodes: %d\n", subnodes->numfields );
#endif
      int jdx;
      for( jdx=0; jdx<subnodes->numfields; jdx++ ) {
        subnode = (EQstruct*) subnodes->fields[jdx].value;
        XMLnode2MatlabStruct( subnode, Msubnode );
      }  
      setField( Msubnode, field->elementname, struct_tmp );    
    }
    else {
      setField( field, struct_tmp );
    }
  }


  if ( localname == NULL ) {
    localname = (char**)mxMalloc( sizeof(char*) );
    strcpy( localname[0], "Unknown");
  }

  setField( struct_tmp, localname[0], Mstruct );

  return;
}





