/**======================================================================
     Gateway routine to read in the hsx structure generated by the SIESTA package.
     Copyright (C) 2016 Peter Rakyta, Ph.D.

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see http://www.gnu.org/licenses/.
     
======================================================================*/
#include "mex.h"
#include <unistd.h>
#include <string.h>
#include "EQuUs_MATLAB.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __INTEL_COMPILER
void hsx_equus_mp_read_hsx_( char* fname, long *fname_len, EQstruct* eq_struct);
#else
void __hsx_equus_MOD_read_hsx( char* fname, long *fname_len, EQstruct* eq_struct);
#endif  


#ifdef __cplusplus
 }
#endif




/** Gateway routine to read in the hsx structure generated by the SIESTA package (see http://www.mathworks.com/help/matlab/matlab_external/standalone-example.html#zmw57dd0e18558).
 \param nlhs	Number of output (left-side) arguments, or the size of the plhs array.
 \param plhs	Array of output arguments.
 \param nrhs	Number of input (right-side) arguments, or the size of the prhs array.
 \param prhs	Array of input arguments.
*/
void mexFunction(int nlhs, mxArray *plhs[],
                 int nrhs, const mxArray *prhs[])
{
      /* input filename */
      char *filename;
      /* length of the filename */
      long buflen;
      /* field names of the output structure */
      char **fieldnames;  



/*-----------------------------------------------------------------------*/
/*     Check for proper number of arguments. */
      if(nrhs != 1) {
         mexErrMsgIdAndTxt ("siesta_EQuUs:hsx_m:nInput", "One input required.");
      }
      else if (nlhs > 1) {
         mexErrMsgIdAndTxt ("siesta_EQuUs:hsx_m:nOutput", "Too many output arguments.");
      }

/*     Validate inputs */
      
/*     Check that the input is a double.*/
      if ((mxIsChar(prhs[0]) != 1)) {
         mexErrMsgIdAndTxt ("siesta_EQuUs:hsx_m:NonString", "Input must be a string.");
      }

      /* get the length of the input string */
      buflen = (mxGetM(prhs[0]) * mxGetN(prhs[0])) + 1;


      /* copy the string data from prhs[0] into a C string.    */
      filename = mxArrayToString(prhs[0]);
#ifdef DEBUG
      mexPrintf( "%s\n", filename );
#endif

      if (access(filename, F_OK) == -1) {
        mexErrMsgIdAndTxt ("siesta_EQuUs:hsx_m:FileNotExist", "The given file %s does not exist.", filename);
      }

      EQelement eq_field, *eq_fields;
      EQstruct eq_hsx;


      /* Retriving the HSX structure */
#ifdef __INTEL_COMPILER
      hsx_equus_mp_read_hsx_( filename, &buflen, &eq_hsx);  
#else
       __hsx_equus_MOD_read_hsx( filename, &buflen, &eq_hsx);
#endif  

      /* obtaining the fieldnames */
      fieldnames =  getFieldnames( &eq_hsx );

      /* creating the MATLAB version of the output structure */
      plhs[0] = mxCreateStructMatrix(1, 1, eq_hsx.numfields, (const char**)fieldnames);

#ifdef DEBUG
      mexPrintf( "\nSetting the fields\n*********************************\n" );
#endif
      int idx;
      for( idx=0; idx<eq_hsx.numfields; idx++ ) {
#ifdef DEBUG
        mexPrintf( "setting field: %s:\n", fieldnames[idx] );
#endif
        setField( &eq_hsx.fields[idx], plhs[0] );
    }

}



