%    Automatic test procedure - based on EQuUs v4.8
%    Copyright (C) 2015 Peter Rakyta, Ph.D.
%
%    This program is free software: you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation, either version 3 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%
%    You should have received a copy of the GNU General Public License
%    along with this program.  If not, see http://www.gnu.org/licenses/.
%
%> @file automatic_test.m
%> @brief Launches the automatic test procedure. First, source code functionalities are checked by processing the unit tests in directory 'UnitTests' and then larger computational tasks are performed via examples in directory 'Examples'
%> @Available
%> EQuUs v4.9 or later
%
%> @brief Launches the automatic test procedure. First, source code functionalities are checked by processing the unit tests in directory 'UnitTests' and then larger computational tasks are performed via examples in directory 'Examples'
function automatic_test()

filename = mfilename('fullpath');
root = fileparts( filename );
cd(root)

         
         
%% performing larger computational tasks        
         
% Examples to run
fncnames = { 'magnetic_ribbon', ...
             'MinimalConductivity', ...
             'Graphene_Conductivity', ...
             'magnetic_barrier', ...
             'Differential_conductance', ...
             'Graphene_QAD_eig', ...
             'Graphene_antidot_transport', ...
             'JosephsonCurrent', ...
             'SpectralDensity', ...
             'SSH_transport', ...
             'SSH_Josephson' ...
             };
         
Dirs     = { 'magnetic_ribbon', ... 1
             'Graphene_Minimal_Conductivity', ... 2
             'Graphene_Conductivity', ... 3
             'magnetic_barrier_0B0', ... 4
             'Differential_conductance', ... 5
             'Graphene_antidot_eigenstate', ... 6
             'Graphene_antidot_transport', ... 7
             'SNS_Josephson', ... 8
             'SpectralDensity', ... 9
             'SSH_transport', ... 10
             'SSH_Josephson' ... 11
             };
         
% These tests are skipped during the test procedure
skip_tests = [10 11];
             
% the filename of the output to be exported in          
output_file = 'automatic_test_results.txt';



% removing previous debug files
delete( [root, filesep, 'Examples', filesep, '*', filesep, 'debug.txt'] );

test_results = false(size(fncnames));

for jdx = 1:length(fncnames)
    
    if ~isempty( find(jdx == skip_tests, 1) )
        continue
    end
    
    test_results(jdx) = runTest( fncnames{jdx}, Dirs{jdx});
    
    if test_results(jdx)
        fid = fopen(output_file, 'a');
        fprintf(fid, ['Test ', fncnames{jdx}, ' finished succesfully. Check the results.\n']);
        fclose(fid);        
    else
        fid = fopen(output_file, 'a');
        fprintf(fid, ['Test ', fncnames{jdx}, ' failed due to coding error.\n']);
        fclose(fid);  
    end
end




if norm( true(size(test_results)) - test_results ) <= 1e-6
    disp('*************************');
    disp('*************************');
    disp('All tests passed.')
    disp('*************************');
else
    for jdx = 1:length(fncnames)
        if test_results( jdx )
            continue
        end
        failed_fncname = fncnames{ jdx }; 
        failed_Dir = Dirs{jdx};
        disp(['Test ', fullfile(root, 'Examples', failed_Dir, failed_fncname), ' failed.']);
    end
end




%% runTest
%> @brief Runs the test for a specific example
%> @param fncname The name of the exmaple to be started.
%> @param directory The absolut path to the directory containing the test case.
    function ret = runTest( fncname, directory)
        
        disp( '**************************************' )
        disp( ['starting test: ', fncname] )
        disp( '**************************************' )
        try
            directory = fullfile( root, 'Examples', directory);
            filenum = getFilenum( directory );
            cd(directory)
            feval( fncname, filenum );
            cd(root)
            ret = true;
        catch err
            ret = false;
            return;
        end
            
    end


%% getFilenum
%> @brief Determines the up-comming 'filenum' input.
%> @param directory The absolut path to the directory containing the test case.
%> @return Returns with the next filenum value to run the test case.
    function ret = getFilenum( directory )
        ret = 1;
        
        numerics = {'1' '2' '3' '4' '5' '6' '7' '8' '9' '0'};
        
        start_idx = [];
        end_idx = [];
        
        try
            list = dir( fullfile( directory, 'results', '*.mat') );
        catch err
            ret = 1;
            return
        end
        
        if isempty( list )
            ret = 1;
            return
        end
        
        
        for kdx = 1:length(list)
            filename_tmp = list(kdx).name;
            for idx = 1:length(filename_tmp)
                character = filename_tmp(idx);
                comp = find(strcmp( character, numerics ));
            
                if norm(comp) > 0
                    if isempty( start_idx )
                        start_idx = idx;
                    end
                else
                    if isempty( end_idx ) && ~isempty( start_idx )
                        end_idx = idx;
                    end
                end
            
            
                if ~isempty( start_idx ) && ~isempty(end_idx )
                    number_tmp = str2double( filename_tmp(start_idx:end_idx-1) );
                    if ret < number_tmp
                        ret = number_tmp;
                    end
                    start_idx = [];
                    end_idx   = [];
                end
            
            
            end
        end
        
        ret = ret+1;
        
        
        
    end


end
